#pragma once

#include "util/main.cpp"

#include "bbi_headers.cpp"
#include "bbi_locs.cpp"
#include "bbi_data_tree.cpp"
#include "bbi_data_values.cpp"
#include "bbi_reader.cpp"

#include "bam_header.cpp"
#include "bam_index.cpp"
#include "bam_entry.cpp"
#include "bam_reader.cpp"

#include "hic_headers.cpp"
#include "hic_locs.cpp"
#include "hic_matrices_metadata.cpp"
#include "hic_matrices_data.cpp"
#include "hic_reader.cpp"

#include "genomes.cpp"
